import numpy as np
from PIL import Image, ImageDraw, ImageEnhance, ImageFont

from src.Base.Century import env as _env
from env import SHORT_PATH

IMG_PATH = SHORT_PATH + "src/Base/Century/images/"
BG_SIZE_BASE = np.array((2400, 1600))
CARD_SIZE_BASE = np.array((150, 200))
RATIO = 1
TOKEN_SIZE = np.array([500, 50]) * RATIO
TOKEN_SIZE_BOARD = np.array([150, 25]) * RATIO


COIN_SIZE = np.array([50, 50]) * RATIO
BG_SIZE = BG_SIZE_BASE * RATIO
CARD_SIZE = CARD_SIZE_BASE * RATIO


class Sprites:
    def __init__(self) -> None:
        self.background = Image.open(IMG_PATH + "back_ground.jpg").resize(BG_SIZE)
        self.myFont = ImageFont.truetype(
            "src/ImageFonts/FreeMonoBoldOblique.ttf", size=50 * RATIO
        )
        self.myFont_end = ImageFont.truetype(
            "src/ImageFonts/FreeMonoBoldOblique.ttf", size=25 * RATIO
        )

        self.font2 = ImageFont.truetype(
            "src/ImageFonts/FreeMonoBoldOblique.ttf", size=30 * RATIO
        )
        self.normal_card = []
        normal_card = np.arange(45)
        for i in range(len(normal_card)):
            self.normal_card.append(
                Image.open(f"./src/Base/Century/images/{i}.png").resize(CARD_SIZE)
            )

        self.victory_card = []
        victory_card = np.arange(36)
        for i in range(len(victory_card)):
            self.victory_card.append(
                Image.open(f"./src/Base/Century/images/victory_{i}.png").resize(
                    CARD_SIZE
                )
            )

        self.ALL_CARD_POINT_IN4 = np.array(
            [
                [0, 0, 0, 5, 200],
                [0, 0, 2, 3, 180],
                [0, 0, 3, 2, 170],
                [0, 0, 0, 4, 160],
                [0, 2, 0, 3, 160],
                [0, 0, 5, 0, 150],
                [0, 0, 2, 2, 140],
                [0, 3, 0, 2, 140],
                [2, 0, 0, 3, 140],
                [0, 2, 3, 0, 130],
                [0, 0, 4, 0, 120],
                [0, 2, 0, 2, 120],
                [0, 3, 2, 0, 120],
                [2, 2, 0, 0, 60],
                [3, 2, 0, 0, 70],
                [2, 3, 0, 0, 80],
                [2, 0, 2, 0, 80],
                [0, 4, 0, 0, 80],
                [3, 0, 2, 0, 90],
                [2, 0, 0, 2, 100],
                [0, 5, 0, 0, 100],
                [0, 2, 2, 0, 100],
                [2, 0, 3, 0, 110],
                [3, 0, 0, 2, 110],
                [1, 1, 1, 3, 200],
                [0, 2, 2, 2, 190],
                [1, 1, 3, 1, 180],
                [2, 0, 2, 2, 170],
                [1, 3, 1, 1, 160],
                [2, 2, 0, 2, 150],
                [3, 1, 1, 1, 140],
                [2, 2, 2, 0, 130],
                [0, 2, 1, 1, 120],
                [1, 0, 2, 1, 120],
                [1, 1, 1, 1, 120],
                [2, 1, 0, 1, 90],
            ]
        )

        self.ALL_CARD_IN4 = np.array(
            [
                [0, 0, 0, 0, 2, 0, 0, 0, 0],
                [0, 0, 0, 0, 3, 0, 0, 0, 0],
                [0, 0, 0, 0, 4, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 1, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 1, 0, 0],
                [0, 0, 0, 0, 2, 1, 0, 0, 0],
                [0, 0, 0, 0, 0, 2, 0, 0, 0],
                [0, 0, 0, 0, 1, 0, 1, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 1, 0],
                [2, 0, 0, 0, 0, 2, 0, 0, 0],
                [2, 0, 0, 0, 0, 0, 1, 0, 0],
                [3, 0, 0, 0, 0, 0, 0, 1, 0],
                [3, 0, 0, 0, 0, 3, 0, 0, 0],
                [3, 0, 0, 0, 0, 1, 1, 0, 0],
                [4, 0, 0, 0, 0, 0, 2, 0, 0],
                [4, 0, 0, 0, 0, 0, 1, 1, 0],
                [5, 0, 0, 0, 0, 0, 0, 2, 0],
                [5, 0, 0, 0, 0, 0, 3, 0, 0],
                [0, 1, 0, 0, 3, 0, 0, 0, 0],
                [0, 2, 0, 0, 0, 0, 2, 0, 0],
                [0, 2, 0, 0, 3, 0, 1, 0, 0],
                [0, 2, 0, 0, 2, 0, 0, 1, 0],
                [0, 3, 0, 0, 0, 0, 3, 0, 0],
                [0, 3, 0, 0, 0, 0, 0, 2, 0],
                [0, 3, 0, 0, 1, 0, 1, 1, 0],
                [0, 3, 0, 0, 2, 0, 2, 0, 0],
                [0, 0, 1, 0, 4, 1, 0, 0, 0],
                [0, 0, 1, 0, 1, 2, 0, 0, 0],
                [0, 0, 1, 0, 0, 2, 0, 0, 0],
                [0, 0, 2, 0, 2, 1, 0, 1, 0],
                [0, 0, 2, 0, 0, 0, 0, 2, 0],
                [0, 0, 2, 0, 2, 3, 0, 0, 0],
                [0, 0, 2, 0, 0, 2, 0, 1, 0],
                [0, 0, 3, 0, 0, 0, 0, 3, 0],
                [0, 0, 0, 1, 0, 0, 2, 0, 0],
                [0, 0, 0, 1, 3, 0, 1, 0, 0],
                [0, 0, 0, 1, 0, 3, 0, 0, 0],
                [0, 0, 0, 1, 2, 2, 0, 0, 0],
                [0, 0, 0, 1, 1, 1, 1, 0, 0],
                [0, 0, 0, 2, 1, 1, 3, 0, 0],
                [0, 0, 0, 2, 0, 3, 2, 0, 0],
                [1, 1, 0, 0, 0, 0, 0, 1, 0],
                [2, 0, 1, 0, 0, 0, 0, 2, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 2],
                [0, 0, 0, 0, 0, 0, 0, 0, 3],
                [0, 0, 0, 0, 0, 0, 0, 0, 0],
            ]
        )

        self.coin_gold = Image.open(f"./src/Base/Century/images/coin-gold.png").resize(
            COIN_SIZE
        )
        self.coin_silver = Image.open(
            f"./src/Base/Century/images/coin-silver.png"
        ).resize(COIN_SIZE)
        self.token = Image.open(f"./src/Base/Century/images/token.png").resize(
            TOKEN_SIZE
        )
        self.token_board = Image.open(f"./src/Base/Century/images/token.png").resize(
            TOKEN_SIZE_BOARD
        )

        self.background.paste(self.token, (1600 * RATIO, 450 * RATIO))
        self.background.paste(self.token, (1600 * RATIO, 675 * RATIO))
        self.background.paste(self.token, (1600 * RATIO, 900 * RATIO))
        self.background.paste(self.token, (1600 * RATIO, 1125 * RATIO))
        self.background.paste(self.token, (1600 * RATIO, 1350 * RATIO))
        self.player_text_coor = [
            (1400 * RATIO, 500 * RATIO),
            (1400 * RATIO, 725 * RATIO),
            (1400 * RATIO, 950 * RATIO),
            (1400 * RATIO, 1175 * RATIO),
            (1400 * RATIO, 1400 * RATIO),
        ]


sprites = Sprites()


def get_description(action):
    if action < 0 or action >= _env.getActionSize():
        return ""
    else:
        action_mean = _env.ACTIONS_MEAN[action]

    return f"player {action_mean}"


class Env_components:
    def __init__(self, env, winner, list_other) -> None:
        self.env = env
        self.winner = winner
        self.list_other = list_other


def get_env_components():
    env = _env.initEnv()
    winner = -1
    list_other = np.array([-1, 1, 2, 3, 4])
    np.random.shuffle(list_other)
    env_components = Env_components(env, winner, list_other)
    return env_components


def get_main_player_state(
    env_components: Env_components, list_agent, list_data, action=None
):
    if not action is None:
        env_components.env = _env.stepEnv(env_components.env, action)

    my_idx = np.where(env_components.list_other == -1)[0][0]
    env_components.winner = -1

    while _env.system_check_end(env_components.env):
        p_idx = int(env_components.env[_env.ENV_ID_ACTION])
        state = _env.getAgentState(env_components.env)

        if env_components.list_other[p_idx] == -1:
            win = -1
            return win, state, env_components

        agent = list_agent[env_components.list_other[p_idx] - 1]
        data = list_data[env_components.list_other[p_idx] - 1]
        # print('QUA')
        action, data = agent(state, data)

        env_components.env = _env.stepEnv(env_components.env, action)

    env_components.env[_env.ENV_CHECK_END] = 1
    env_components.winner = _env.check_winner(env_components.env)
    for p_idx in range(5):
        if p_idx != my_idx:
            env_components.env[_env.ENV_ID_ACTION] = p_idx
            state = _env.getAgentState(env_components.env)

            agent = list_agent[env_components.list_other[p_idx] - 1]
            data = list_data[env_components.list_other[p_idx] - 1]
            action, data = agent(state, data)

    env_components.env[_env.ENV_ID_ACTION] = my_idx
    state = _env.getAgentState(env_components.env)
    # print(env_components.winner)
    if my_idx == env_components.winner:
        win = 1
    else:
        win = 0

    return win, state, env_components


def get_state_image(state=None):
    background = sprites.background.copy()
    background = draw_state(state, background)
    return background


def draw_state(state, back_ground):
    phase = np.where(state[267:272])[0][0] + 1
    check_end = int(state[266])
    draw = ImageDraw.Draw(back_ground)

    # vẽ thẻ thường chưa dùng
    normal_card = np.where(state[6:51])[0]
    if len(normal_card) > 0:
        for id in range(len(normal_card)):
            back_ground.paste(
                sprites.normal_card[normal_card[id]], (40 * id * RATIO, 5 * RATIO)
            )

    # vẽ thẻ thường đã dùng
    normal_card_down = np.where(state[51:96])[0]
    if len(normal_card_down) > 0:
        for id in range(len(normal_card_down)):
            back_ground.paste(
                sprites.normal_card[normal_card_down[id]],
                (40 * id * RATIO, 225 * RATIO),
            )
    # vẽ thẻ thường trên bàn
    card_on_board = state[120:174]
    token_on_board = state[174:194].astype(np.int64)
    for id in range(6):
        card_id_in4 = card_on_board[9 * id : 9 * (id + 1)]
        if np.sum(card_id_in4) == 0:
            break
        id_img = -1
        for idd in range(len(sprites.ALL_CARD_IN4)):
            if np.sum(card_id_in4 == sprites.ALL_CARD_IN4[idd]) == 9:
                id_img = idd
                break
        back_ground.paste(
            sprites.normal_card[id_img], (175 * id * RATIO + 50 * RATIO, 825 * RATIO)
        )
        if id < 5:
            token_free = token_on_board[4 * id : 4 * (id + 1)]
            if np.sum(token_free) > 0:
                draw.text(
                    (175 * id * RATIO + 50 * RATIO, 750 * RATIO),
                    f"{token_free[0]}  {token_free[1]}  {token_free[2]}  {token_free[3]}",
                    (255, 250, 0),
                    font=sprites.myFont_end,
                )
                back_ground.paste(
                    sprites.token_board, (175 * id * RATIO + 50 * RATIO, 775 * RATIO)
                )

    card_point_on_board = state[194:219]
    for id in range(5):
        card_id_in4 = card_point_on_board[5 * id : 5 * (id + 1)]
        id_img = -1
        for idd in range(len(sprites.ALL_CARD_POINT_IN4)):
            if np.sum(card_id_in4 == sprites.ALL_CARD_POINT_IN4[idd]) == 5:
                id_img = idd
                break
        back_ground.paste(
            sprites.victory_card[id_img], (175 * id * RATIO + 50 * RATIO, 1100 * RATIO)
        )

    draw.text(
        (25 * RATIO, 500 * RATIO),
        f"Phase: {phase}\nCheck_end: {check_end}  ",
        (255, 0, 0),
        font=sprites.myFont,
    )
    if state[220]:
        back_ground.paste(
            sprites.coin_gold, (100 * RATIO, 1300 * RATIO), sprites.coin_gold
        )
        draw.text(
            (100 * RATIO, 1350 * RATIO),
            f"{int(state[220])}",
            (255, 255, 0),
            font=sprites.myFont,
        )
    if state[219]:
        back_ground.paste(
            sprites.coin_silver, (250 * RATIO, 1300 * RATIO), sprites.coin_silver
        )
        draw.text(
            (250 * RATIO, 1350 * RATIO),
            f"{int(state[219])}",
            (255, 255, 255),
            font=sprites.myFont,
        )

    player_in4 = state[:6].astype(np.int64)
    other_player_in4 = state[96:120].astype(np.int64)
    id_start = state[272:277].astype(np.int64)

    draw.text(
        sprites.player_text_coor[0],
        f"Token: {player_in4[2]}   {player_in4[3]}   {player_in4[4]}   {player_in4[5]}\nScore: {player_in4[0]}, {id_start[0]} \nNumber_point_card: {player_in4[1]} ",
        (255, 255, 0),
        font=sprites.myFont,
    )

    for i in range(4):
        player_in4_i = other_player_in4[6 * i : 6 * (i + 1)]
        draw.text(
            sprites.player_text_coor[i + 1],
            f"Token: {player_in4_i[2]}   {player_in4_i[3]}   {player_in4_i[4]}   {player_in4_i[5]}\nScore: {player_in4_i[0]}, {id_start[i+1]} \nNumber_point_card: {player_in4_i[1]} ",
            (255, 255, 0),
            font=sprites.myFont,
        )
    # state[240] = 1
    if np.sum(state[221:266]) > 0:
        last_action = np.where(state[221:266] == 1)[0][0]
        # print('check', last_action)
        back_ground.paste(sprites.normal_card[last_action], (500 * RATIO, 1400 * RATIO))

    return back_ground


def draw_card(card_back_ground, card_in4, list_img_source):
    give = card_in4[0:4]
    receive = card_in4[4:8]
    upgrade = card_in4[-1]
    # print(give, receive, upgrade)
    len_card = np.sum(card_in4)
    if upgrade != 0 or np.sum(give) != 0:
        len_card += 1
    img = Image.new("RGB", (50, 40 * len_card), color="white")
    if upgrade != 0:
        cube = list_img_source[-1]
        for i in range(upgrade):
            idx = np.array([5, 40 * i])
            img.paste(cube, tuple(idx))
        up_img = Image.open("./src/Base/Century/images/uparrow.png")
        img.paste(up_img, (5, 40 * upgrade))
    else:
        if np.sum(give) != 0:
            # print('vào đây')
            card_back_ground = Image.open(
                f"./src/Base/Century/images/img/resource-card-trading.png"
            ).resize((300, 400))
            down_img = Image.open("./src/Base/Century/images/downarrow.png")
            index_run = 0
            for type in range(4):
                if give[type] != 0:
                    cube = list_img_source[type]
                    for i in range(give[type]):
                        idx = np.array([5, 40 * index_run])
                        img.paste(cube, tuple(idx))
                        index_run += 1
            img.paste(down_img, (5, 40 * index_run))
            index_run += 1
            # print('check', index_run)
            for type in range(4):
                if receive[type] != 0:
                    cube = list_img_source[type]
                    for i in range(receive[type]):
                        idx = np.array([5, 40 * index_run])
                        img.paste(cube, tuple(idx))
                        index_run += 1
        else:
            # print('vào đây2')

            card_back_ground = Image.open(
                f"./src/Base/Century/images/img/resource-card-production.png"
            ).resize((300, 400))
            index_run = 0
            for type in range(4):
                if receive[type] != 0:
                    cube = list_img_source[type]
                    for i in range(receive[type]):
                        idx = np.array([5, 40 * index_run])
                        img.paste(cube, tuple(idx))
                        index_run += 1

    card_back_ground.paste(img, (30, 20))
    return card_back_ground


def draw_point_card(card_back_ground, card_in4, list_img_source):
    give = card_in4[:4]
    score = card_in4[-1]
    font_t = ImageFont.truetype("src/ImageFonts/FreeMonoBoldOblique.ttf", size=100)
    len_card = np.sum(give)
    img = Image.new("RGB", (40 * len_card, 50), color="white")
    index_run = 0
    # print(give)
    for type in range(4):
        if give[type] != 0:
            cube = list_img_source[type]
            for i in range(give[type]):
                idx = np.array([40 * index_run, 5])
                img.paste(cube, tuple(idx))
                index_run += 1
    draw = ImageDraw.Draw(card_back_ground)
    draw.text((50, 200), f"{score}", (255, 255, 0), font=font_t)
    card_back_ground.paste(img, (int((300 - 40 * len_card) / 2), 330))
    return card_back_ground


# type = ['yellow', 'red', 'green', 'brown', 'any']

# list_type = []
# for type_i in type:
#     list_type.append(Image.open(f'./src/Base/Century/images/img/cube-{type_i}.png'))

# list_type



import sys
import numpy as np
from numba import njit
import env

game_name = sys.argv[1]
env.make(game_name)


from numba.typed import List

MAX_DEPTH = 7


def DataAgent():
    perData = List()

    perData.append(np.array([[0.0]], dtype=np.float64))  #  0: Đếm số trận đấu đã train
    perData.append(
        np.full((2, env.getActionSize()), 1e-9, dtype=np.float64)
    )  #  Bias và temp Bias
    perData[1][1] = np.arange(1, env.getActionSize() + 1, dtype=np.float64)
    np.random.shuffle(perData[1][1])

    perData.append(
        np.full((2, env.getStateSize()), 0.0, dtype=np.float64)
    )  #  Minimum state và maximum state
    perData[2][0] = 1e9
    perData[2][1] = -1e9

    return perData


@njit
def Train(state, perData):
    if perData[0][0][0] < 10000:
        validActions = env.getValidActions(state)
        bias = perData[1][1] * validActions
        action = np.argmax(bias)

        reward = env.getReward(state)
        if reward != -1:
            perData[0][0][0] += 1
            if reward == 1:
                perData[1][0] += perData[1][1]
            elif reward == 0:
                perData[1][0] -= perData[1][1]

            np.random.shuffle(perData[1][1])
            if perData[0][0][0] == 10000:
                min_ = np.min(perData[1][0])
                if min_ < 0:
                    perData[1][0] += 1.0 - min_

                perData[1][0] /= np.max(perData[1][0])
                perData[1][0] **= 2

        return action, perData

    elif perData[0][0][0] < 20000 and len(perData) < 11:
        perData[2][0] = np.minimum(perData[2][0], state)
        perData[2][1] = np.maximum(perData[2][1], state)

        validActions = env.getValidActions(state)
        actions = np.where(validActions == 1)[0]
        a_idx = np.random.randint(0, actions.shape[0])

        reward = env.getReward(state)
        if reward != -1:
            perData[0][0][0] += 1
            if perData[0][0][0] == 20000:
                size_ = int(np.round(np.max(perData[2][1] - perData[2][0])))
                if len(perData) < 11:
                    perData.append(
                        np.full((env.getStateSize(), size_), 0.0, dtype=np.float64)
                    )  #  3 win_state
                    perData.append(
                        np.full((env.getStateSize(), size_), 1e-9, dtype=np.float64)
                    )  #  4 last_state
                    perData.append(
                        np.full(
                            (env.getActionSize(), env.getStateSize()),
                            0.0,
                            dtype=np.float64,
                        )
                    )  #  5 Delta_state
                    perData.append(
                        np.full(
                            (env.getActionSize(), env.getStateSize()),
                            0.0,
                            dtype=np.float64,
                        )
                    )  #  6 Pre_state
                    perData.append(
                        np.full((env.getActionSize(), 1), 0.0, dtype=np.float64)
                    )  #  7 Số lần xuất hiện
                    perData.append(
                        np.array([[-1.0]], dtype=np.float64)
                    )  #  8 Pre_action
                    perData.append(
                        np.full((env.getStateSize(), size_), 0.01, dtype=np.float64)
                    )  #  9 = 3 / 4
                    perData.append(
                        np.full(
                            (env.getActionSize(), env.getStateSize()),
                            0.0,
                            dtype=np.float64,
                        )
                    )  #  10 = 5 / 7

        return actions[a_idx], perData

    elif perData[0][0][0] < 100000:
        if perData[8][0][0] > -1.0:
            perData[5][int(perData[8][0][0])] += (
                state - perData[6][int(perData[8][0][0])]
            )
            perData[7][0][int(perData[8][0][0])] += 1

        validActions = env.getValidActions(state)
        actions = np.where(validActions == 1)[0]
        a_idx = np.random.randint(0, actions.shape[0])

        perData[8][0][0] = actions[a_idx]
        perData[6][actions[a_idx]] = state

        reward = env.getReward(state)
        if reward != -1:
            perData[0][0][0] += 1
            perData[8][0][0] = -1.0
            arr = np.where(state < perData[2][0], perData[2][0], state)
            arr = np.where(arr > perData[2][1], perData[2][1], arr)
            arr = np.round(arr, 0, arr).astype(np.int64)
            for i in range(arr.shape[0]):
                perData[4][i][arr[i]] += 1

            if reward == 1:
                for i in range(arr.shape[0]):
                    perData[3][i][arr[i]] += 1

            if perData[0][0][0] == 100000:
                perData[0][0][0] = 0
                perData[9] = perData[3] / perData[4]
                perData[10] = perData[5] / perData[7]

        return actions[a_idx], perData


@njit
def get_value_state(state, value):
    score = 0.0
    state = state.astype(np.int64)
    for i in range(state.shape[0]):
        score += value[i][state[i]]

    return score / state.shape[0]


@njit
def weighted_random(p: np.ndarray):
    a = np.sum(p)
    b = np.random.uniform(0, a)
    for i in range(len(p)):
        b -= p[i]
        if b <= 0:
            return i

    return 999999


@njit
def scoring(action, state, depth, delta, bias, score):
    new_state = state + delta[action]
    new_state_round = np.empty_like(new_state)
    new_state_round = np.round(new_state, 0, new_state_round)
    validActions = env.getValidActions(new_state_round)
    actions = np.where(validActions == 1)[0]
    reward = env.getReward(new_state_round)

    if depth >= MAX_DEPTH:
        return get_value_state(new_state_round, score)

    if reward == 1:
        return (MAX_DEPTH - depth + 1) / (MAX_DEPTH + 1) + get_value_state(
            new_state_round, score
        )

    if reward == 0:
        return (depth - MAX_DEPTH - 1) / (MAX_DEPTH + 1) + get_value_state(
            new_state_round, score
        )

    if actions.shape[0] == 0:
        return (depth - MAX_DEPTH) / (MAX_DEPTH + 1) + get_value_state(
            new_state_round, score
        )

    temp = bias[actions]
    a_idx = weighted_random(temp)
    depth_ = depth + 1
    return scoring(actions[a_idx], new_state, depth_, delta, bias, score)


@njit
def Test(state, perData):
    validActions = env.getValidActions(state)
    actions = np.where(validActions == 1)[0]
    score = np.full(actions.shape, 0.0, dtype=np.float64)
    for i in range(actions.shape[0]):
        score[i] = scoring(actions[i], state, 0, perData[0], perData[1][0], perData[2])

    a_idx = np.argmax(score)
    return actions[a_idx], perData


def convert_to_save(perData):
    data = List()
    data.append(perData[10])
    min_ = np.min(perData[1][0])
    if min_ < 0:
        perData[1][0] += 1.0 - min_

    perData[1][0] /= np.max(perData[1][0])
    perData[1][0] **= 2
    data.append(perData[1])
    data.append(perData[9])
    return data


def convert_to_test(perData):
    return List(perData)



import numpy as np
from numba import njit

from src.Base.Sheriff.docs.index import *

# from index import*


@njit()
def getActionSize():
    return 82


@njit()
def getAgentSize():
    return 4


@njit()
def getStateSize():
    return 478


@njit()
def system_check_end(env_state):
    if env_state[ENV_ROUND] <= 8:
        return True
    else:
        return False


@njit()
def initEnv():
    env_state = np.zeros(ENV_LENGTH)
    normal_card = NORMAL_CARD.copy()
    royal_card = ROYAL_CARD.copy()

    np.random.shuffle(normal_card)
    for id in range(NUMBER_PLAYER):
        player_i = START_PLAYER.copy()
        if id == 0:
            player_i[2] = 1
            card_player = player_i[-15:]
            for card in normal_card[id * 5 : (id + 1) * 5]:
                card_player[card - 1] += 1
            player_i[-15:] = card_player
            env_state[ENV_PLAYER_IN4:ATTRIBUTE_PLAYER] = player_i
        else:
            card_player = player_i[-15:]
            for card in normal_card[id * 5 : (id + 1) * 5]:
                card_player[card - 1] += 1
            player_i[-15:] = card_player
            env_state[ATTRIBUTE_PLAYER * id : ATTRIBUTE_PLAYER * (id + 1)] = player_i
    # print(env_state[ENV_PLAYER_IN4 : ATTRIBUTE_PLAYER*4])
    all_card = np.concatenate((normal_card[20:], royal_card))
    np.random.shuffle(all_card)
    env_state[ENV_LEFT_CARD : ENV_LEFT_CARD + CARD_OPEN_START] = all_card[:5]
    env_state[ENV_RIGHT_CARD : ENV_RIGHT_CARD + CARD_OPEN_START] = all_card[5:10]
    env_state[ENV_DOWN_CARD : ENV_DOWN_CARD + NUMBER_CARD_USE] = all_card[10:]
    env_state[ENV_ID_ACTION] = 1
    env_state[ENV_ROUND] = 1
    env_state[ENV_PHASE] = 1
    return env_state


@njit()
def getAgentState(env_state):
    player_action = int(env_state[ENV_ID_ACTION])
    player_state = np.zeros(P_LENGTH, dtype=np.float64)
    player_state[P_PLAYER_IN4:P_OTHER_PLAYER_IN4] = env_state[
        ATTRIBUTE_PLAYER * player_action : ATTRIBUTE_PLAYER * (player_action + 1)
    ]
    number_card_in_bag = np.zeros(NUMBER_PLAYER)
    number_card_in_bag[0] = np.sum(
        env_state[
            ATTRIBUTE_PLAYER * player_action : ATTRIBUTE_PLAYER * (player_action + 1)
        ][-45:-30]
    )

    # thông tin các người chơi khác
    for idx in range(1, NUMBER_PLAYER):
        id = (player_action + idx) % NUMBER_PLAYER
        all_other_player_in4 = env_state[
            ATTRIBUTE_PLAYER * id : ATTRIBUTE_PLAYER * (id + 1)
        ]
        other_player_in4 = all_other_player_in4[:OTHER_ATTRIBUTE_PLAYER]
        other_player_drop_card = env_state[
            ENV_TEMP_DROP
            + NUMBER_TYPE_CARD * id : ENV_TEMP_DROP
            + NUMBER_TYPE_CARD * (id + 1)
        ]
        player_state[
            P_OTHER_PLAYER_IN4
            + OTHER_ATTRIBUTE_PLAYER * (idx - 1) : P_OTHER_PLAYER_IN4
            + OTHER_ATTRIBUTE_PLAYER * idx
        ] = other_player_in4
        # thẻ người chơi bỏ trong túi
        number_card_in_bag[idx] = np.sum(all_other_player_in4[-45:-30])
        # thẻ người chơi khác đã bỏ
        player_state[
            P_CARD_OTHER_PLAYER_DROP
            + NUMBER_TYPE_CARD * (idx - 1) : P_CARD_OTHER_PLAYER_DROP
            + NUMBER_TYPE_CARD * idx
        ] = other_player_drop_card
        # thẻ người chơi có(chỉ hiện khiheets game)
        if env_state[ENV_CHECK_END] == 1:
            player_state[
                P_OTHER_PLAYER_DONE_CARD
                + NUMBER_TYPE_CARD * (idx - 1) : P_OTHER_PLAYER_DONE_CARD
                + NUMBER_TYPE_CARD * idx
            ] = all_other_player_in4[-75:-60]

    # 6 thẻ đầu ở 2 chồng bài
    # trái
    for id in range(MAX_CARD_TAKE):
        card_type_index = int(env_state[ENV_LEFT_CARD + id]) - 1
        if card_type_index >= 0:
            player_state[P_LEFT_CARD + NUMBER_TYPE_CARD * id + card_type_index] = 1
    # Phải
    for id in range(MAX_CARD_TAKE):
        card_type_index = int(env_state[ENV_RIGHT_CARD + id]) - 1
        if card_type_index >= 0:
            player_state[P_RIGHT_CARD + NUMBER_TYPE_CARD * id + card_type_index] = 1
    player_state[P_ROUND] = env_state[ENV_ROUND]
    player_state[P_CHECK_END] = env_state[ENV_CHECK_END]
    player_state[P_PHASE + int(env_state[ENV_PHASE] - 1)] = 1
    player_state[P_ORDER : P_ORDER + NUMBER_PLAYER] = np.concatenate(
        (ALL_ORDER[player_action:], ALL_ORDER[:player_action])
    )

    player_state[
        P_NUMBER_CARD_IN_BAG : P_NUMBER_CARD_IN_BAG + NUMBER_PLAYER
    ] = number_card_in_bag
    # print(player_action, number_card_in_bag)
    return player_state


@njit()
def getValidActions(player_state_origin):
    player_state = player_state_origin.copy()
    list_action_return = np.zeros(82, dtype=np.float64)
    phase_env = np.where(player_state[P_PHASE : P_PHASE + NUMBER_PHASE] == 1)[0][0] + 1
    player_state_own = player_state[:ATTRIBUTE_PLAYER]
    """
        OWNER: 100
            0: player coin
            1: debt
            2: is_police
            [3 : 7]: type_in_bag
            8: coin_bride
            9: number_smuggle
            10: number_card_bride_bag
            -90:-75: card_done_bride
            -75:60: card_done
            -60:-45: card_in_bag_bride
            -45:-30: card_in_bag
            -30:-15: card_bag_drop
            -15: card_hand
    """
    if phase_env == 1:
        list_action_return[15] = 1
        player_card = player_state_own[-15:]
        for id in range(15):
            if player_card[id] == 0:
                continue
            else:
                list_action_return[id] = 1

    elif phase_env == 2:
        # lấy thẻ từ các chồng bài: 16-bài rút, 17-lật trái, 18-lật phải
        list_action_return[16] = 1
        if np.sum(player_state[P_LEFT_CARD : P_LEFT_CARD + NUMBER_TYPE_CARD]) > 0:
            list_action_return[17] = 1
        if np.sum(player_state[P_RIGHT_CARD : P_RIGHT_CARD + NUMBER_TYPE_CARD]) > 0:
            list_action_return[18] = 1

    elif phase_env == 3:
        # trả thẻ bỏ vào  chồng bài lật: 19: lật trái, 20: lật phải
        list_action_return[np.array([19, 20])] = 1

    elif phase_env == 4:
        # lựa chọn dừng bỏ thẻ vào túi đi chợ
        list_action_return[36] = 1
        player_card = player_state_own[-15:]
        for id in range(15):
            if player_card[id] == 0:
                continue
            else:
                list_action_return[id + 21] = 1
        if np.sum(player_state_own[-45:-30]) == 0:
            list_action_return[36] = 0

    elif phase_env == 5:
        # chọn 1 trong 4 loại hàng chính ngạch
        list_action_return[np.array([37, 38, 39, 40])] = 1

    elif phase_env == 6:
        # 41, 42, 43 lần lượt là check người ở vị trí 1,2,3 sau mình
        for id in range(3):
            # kiểm tra xem đã check hay chưa bằng cách xem mặt hàng khai báo
            type_bag_other_player = player_state[
                P_OTHER_PLAYER_IN4
                + OTHER_ATTRIBUTE_PLAYER * id : P_OTHER_PLAYER_IN4
                + OTHER_ATTRIBUTE_PLAYER * (id + 1)
            ][P_TYPE_IN_BAG:P_COIN_BRIBE]
            if np.sum(type_bag_other_player) != 0:
                list_action_return[41 + id] = 1

    elif phase_env == 7:
        # 44 là ko hối lộ nữa, 45 là thêm coin
        if player_state_own[P_COIN] - player_state_own[P_DEBT] > 0:
            list_action_return[np.array([44, 45])] = 1
        else:
            list_action_return[44] = 1

    elif phase_env == 8:
        list_action_return[61] = 1  # action dừng hối lộ thẻ done
        player_card_done = player_state_own[-75:-60]
        for id in range(15):
            if player_card_done[id] == 0:
                continue
            else:
                # bắt đầu duyệt từ việc hối lộ thẻ táo ứng vs action 46
                list_action_return[id + 46] = 1

    elif phase_env == 9:
        list_action_return[77] = 1  # action dừng hối lộ thẻ trong túi
        player_card_bag = player_state_own[-45:-30]
        for id in range(15):
            if player_card_bag[id] == 0:
                continue
            else:
                list_action_return[id + 62] = 1

    elif phase_env == 10:
        # 79 là có check hàng, 80 là cho thoát
        list_action_return[np.array([78, 79])] = 1

    elif phase_env == 11:
        list_action_return[np.array([80, 81])] = 1

    return list_action_return


@njit()
def stepEnv(env_state, action):
    phase_env = env_state[ENV_PHASE]
    id_action = int(env_state[ENV_ID_ACTION])
    player_in4 = env_state[
        ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
    ]

    if phase_env == 1:
        if action != 15:
            player_card = player_in4[-15:]
            temp_drop_card = env_state[
                ENV_TEMP_DROP
                + NUMBER_TYPE_CARD * id_action : ENV_TEMP_DROP
                + NUMBER_TYPE_CARD * (id_action + 1)
            ]
            player_card[action] -= 1
            temp_drop_card[action] += 1
            player_in4[-15:] = player_card
            env_state[
                ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
            ] = player_in4
            env_state[
                ENV_TEMP_DROP
                + NUMBER_TYPE_CARD * id_action : ENV_TEMP_DROP
                + NUMBER_TYPE_CARD * (id_action + 1)
            ] = temp_drop_card
            if np.sum(player_card) == 0:
                env_state[ENV_PHASE] = 2
        else:
            env_state[ENV_PHASE] = 2

    elif phase_env == 2:
        if action == 16:
            player_card = player_in4[-15:]
            # print('check', player_card)
            card_down = env_state[ENV_DOWN_CARD : ENV_DOWN_CARD + NUMBER_CARD_USE]
            number_card_get = int(MAX_CARD_TAKE - np.sum(player_card))
            card_get = card_down[:number_card_get]
            if len(card_get) > 0:
                for card in card_get:
                    player_card[int(card) - 1] += 1
            card_down = np.concatenate(
                (card_down[number_card_get:], np.zeros(number_card_get))
            )
            env_state[ENV_DOWN_CARD : ENV_DOWN_CARD + NUMBER_CARD_USE] = card_down
            player_in4[-15:] = player_card
            env_state[
                ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
            ] = player_in4

            if (
                np.sum(
                    env_state[
                        ENV_TEMP_DROP
                        + NUMBER_TYPE_CARD * id_action : ENV_TEMP_DROP
                        + NUMBER_TYPE_CARD * (id_action + 1)
                    ]
                )
                != 0
            ):
                env_state[ENV_PHASE] = 3
            else:
                # print('chuyển người chơi luôn vì không bỏ thẻ')
                env_state[ENV_PHASE] = 1
                env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4
                if env_state[ENV_ID_ACTION] == (env_state[ENV_ROUND] - 1) % 4:
                    env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4
                    env_state[ENV_PHASE] = 4

        else:
            player_card = player_in4[-15:]
            card_up = env_state[
                ENV_LEFT_CARD
                + NUMBER_CARD_OPEN * (action - 17) : ENV_LEFT_CARD
                + NUMBER_CARD_OPEN * (action - 16)
            ]
            card_get = card_up[0]
            player_card[int(card_get) - 1] += 1
            card_up = np.append(card_up[1:], 0)
            env_state[
                ENV_LEFT_CARD
                + NUMBER_CARD_OPEN * (action - 17) : ENV_LEFT_CARD
                + NUMBER_CARD_OPEN * (action - 16)
            ] = card_up
            player_in4[-15:] = player_card
            env_state[
                ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
            ] = player_in4
            if np.sum(player_card) == 6:
                if (
                    np.sum(
                        env_state[
                            ENV_TEMP_DROP
                            + NUMBER_TYPE_CARD * id_action : ENV_TEMP_DROP
                            + NUMBER_TYPE_CARD * (id_action + 1)
                        ]
                    )
                    != 0
                ):
                    env_state[ENV_PHASE] = 3
                else:
                    # print('chuyển người chơi luôn vì không bỏ thẻ')
                    env_state[ENV_PHASE] = 1
                    env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4
                    if env_state[ENV_ID_ACTION] == (env_state[ENV_ROUND] - 1) % 4:
                        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4
                        env_state[ENV_PHASE] = 4

    elif phase_env == 3:
        temp_card_drop = env_state[
            ENV_TEMP_DROP
            + NUMBER_TYPE_CARD * id_action : ENV_TEMP_DROP
            + NUMBER_TYPE_CARD * (id_action + 1)
        ]
        card_up = env_state[
            ENV_LEFT_CARD
            + NUMBER_CARD_OPEN * (action - 19) : ENV_LEFT_CARD
            + NUMBER_CARD_OPEN * (action - 18)
        ]
        for i in range(len(temp_card_drop)):
            if temp_card_drop[i] > 0:
                card_up = np.append(np.array([i + 1] * int(temp_card_drop[i])), card_up)
        if np.sum(temp_card_drop) > 0:
            card_up = card_up[: int(-np.sum(temp_card_drop))]

        # env_state[ENV_TEMP_DROP + NUMBER_TYPE_CARD * id_action : ENV_TEMP_DROP + NUMBER_TYPE_CARD * (id_action + 1)] = np.zeros(NUMBER_TYPE_CARD)
        env_state[
            ENV_LEFT_CARD
            + NUMBER_CARD_OPEN * (action - 19) : ENV_LEFT_CARD
            + NUMBER_CARD_OPEN * (action - 18)
        ] = card_up
        env_state[ENV_PHASE] = 1
        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4
        if env_state[ENV_ID_ACTION] == (env_state[ENV_ROUND] - 1) % 4:
            env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4
            env_state[ENV_PHASE] = 4

    elif phase_env == 4:
        card_bag = player_in4[-45:-30]
        player_card = player_in4[-15:]
        if action == 36:
            if np.sum(card_bag) == 0:
                print(env_state)
                raise Exception("chưa bỏ thẻ vào túi")
            else:
                env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4
                if env_state[ENV_ID_ACTION] == (env_state[ENV_ROUND] - 1) % 4:
                    env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4
                    env_state[ENV_PHASE] = 5
        else:
            card_bag[action - 21] += 1
            player_card[action - 21] -= 1
            player_in4[-15:] = player_card
            player_in4[-45:-30] = card_bag
            env_state[
                ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
            ] = player_in4
            if np.sum(card_bag) == 5:
                env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4
                if env_state[ENV_ID_ACTION] == (env_state[ENV_ROUND] - 1) % 4:
                    env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4
                    env_state[ENV_PHASE] = 5

    elif phase_env == 5:
        type_bag = action - 36
        player_in4[P_TYPE_IN_BAG:P_COIN_BRIBE][int(type_bag - 1)] = 1
        env_state[
            ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
        ] = player_in4
        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4
        if env_state[ENV_ID_ACTION] == (env_state[ENV_ROUND] - 1) % 4:
            env_state[ENV_PHASE] = 6

    elif phase_env == 6:
        player_checked = action - 40
        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + player_checked) % 4
        env_state[ENV_NUMBER_CHECKED] += 1
        env_state[ENV_PHASE] = 7
        env_state[ENV_LAST_CHECKED + int(env_state[ENV_ID_ACTION])] = 1

    elif phase_env == 7:
        if action == 45:
            player_in4[0] -= 1
            # new: chỉnh vị trí coin hối lộ
            player_in4[P_COIN_BRIBE] += 1
            env_state[
                ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
            ] = player_in4
        else:
            env_state[ENV_PHASE] = 8

    elif phase_env == 8:
        if action == 61:
            env_state[ENV_PHASE] = 9
        else:
            all_card_done = player_in4[-75:-60]
            all_card_bride = player_in4[-90:-75]
            card_bride = action - 46
            all_card_done[card_bride] -= 1
            all_card_bride[card_bride] += 1
            player_in4[-75:-60] = all_card_done
            player_in4[-90:-75] = all_card_bride
            env_state[
                ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
            ] = player_in4

    elif phase_env == 9:
        if action == 77:
            env_state[ENV_PHASE] = 10
            env_state[ENV_ID_ACTION] = (env_state[ENV_ROUND] - 1) % 4
        else:
            # print(player_in4[-90:-60])
            all_card_bag = player_in4[-45:-30]
            all_card_bride_bag = player_in4[-60:-45]
            card_bride = action - 62
            all_card_bag[card_bride] -= 1
            all_card_bride_bag[card_bride] += 1
            # cập nhập số thẻ trong túi dùng hối lộ
            player_in4[P_NUMBER_CARD_BRIBE_BAG] += 1
            player_in4[-45:-30] = all_card_bag
            player_in4[-60:-45] = all_card_bride_bag
            env_state[
                ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
            ] = player_in4

    elif phase_env == 10:
        id_checked = np.where(
            env_state[ENV_LAST_CHECKED : ENV_LAST_CHECKED + NUMBER_PLAYER] == 1
        )[0][0]
        env_state[ATTRIBUTE_PLAYER * id_checked : ATTRIBUTE_PLAYER * (id_checked + 1)]
        player_checked = env_state[
            ATTRIBUTE_PLAYER * id_checked : ATTRIBUTE_PLAYER * (id_checked + 1)
        ]
        type_bag = np.where(player_checked[P_TYPE_IN_BAG:P_COIN_BRIBE] == 1)[0][0] + 1
        if action == 78:
            # có check
            # B1: khôi phục coin
            player_checked[P_COIN] += player_checked[P_COIN_BRIBE]
            player_checked[P_COIN_BRIBE] = 0
            # B3 khôi phục thẻ done
            player_checked[-75:-60] = player_checked[-75:-60] + player_checked[-90:-75]
            player_checked[-90:-75] = np.zeros(15)
            # khôi phục thẻ trong túi
            player_checked[-45:-30] = player_checked[-45:-30] + player_checked[-60:-45]
            player_checked[-60:-45] = np.zeros(15)
            # tính toán cho sheriff
            player_checked_card_bag = player_checked[-45:-30]
            player_checked_card_done = player_checked[-75:-60]
            card_bag_drop = player_in4[-30:-15]
            # cập nhật coin
            if player_checked_card_bag[int(type_bag) - 1] == np.sum(
                player_checked_card_bag
            ):
                penalty = (
                    -player_checked_card_bag[int(type_bag) - 1]
                    * ALL_PENALTY[int(type_bag) - 1]
                )
            else:
                penalty = (
                    np.sum(np.multiply(player_checked_card_bag, ALL_PENALTY))
                    - player_checked_card_bag[int(type_bag) - 1]
                    * ALL_PENALTY[int(type_bag) - 1]
                )
            player_in4[P_COIN] += penalty
            player_checked[P_COIN] -= penalty
            # Kiểm tra xem có đủ tiền ko để ghi nợ
            if player_in4[P_COIN] < 0:
                player_in4[P_DEBT] -= player_in4[P_COIN]
                player_in4[P_COIN] = 0
            if player_checked[P_COIN] < 0:
                player_checked[P_DEBT] -= player_checked[P_COIN]
                player_checked[P_COIN] = 0
            # cập nhật thẻ
            for id in range(15):
                if id == type_bag - 1:
                    player_checked_card_done[id] += player_checked_card_bag[id]
                else:
                    card_bag_drop[id] += player_checked_card_bag[id]
            player_checked[-45:-30] = np.zeros(15)
            player_checked[-75:-60] = player_checked_card_done

            player_checked[P_NUMBER_CARD_BRIBE_BAG] = 0  # số thẻ trong túi dùng hối lộ
            player_checked[
                P_TYPE_IN_BAG:P_COIN_BRIBE
            ] = 0  # loại thẻ người chơi khai báo
            player_checked[P_NUMBER_SMUGGLE] = np.sum(
                player_checked[-75:-60][4:]
            )  # số thẻ buôn lậu thành công của người chơi
            env_state[
                ATTRIBUTE_PLAYER * id_checked : ATTRIBUTE_PLAYER * (id_checked + 1)
            ] = player_checked
            player_in4[-30:-15] = card_bag_drop
            env_state[
                ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
            ] = player_in4
            # cập nhật trong hệ thống
            if np.sum(card_bag_drop) > 0:
                env_state[ENV_PHASE] = 11
                env_state[ENV_ID_ACTION] = (env_state[ENV_ROUND] - 1) % 4
            else:
                env_state[ENV_LAST_CHECKED : ENV_LAST_CHECKED + NUMBER_PLAYER] = 0
                if env_state[ENV_NUMBER_CHECKED] == 3:
                    old_sheriff = int((env_state[ENV_ROUND] - 1) % 4)
                    env_state[
                        ATTRIBUTE_PLAYER * old_sheriff + 2
                    ] = 0  # hủy tư cách sheriff
                    env_state[ENV_PHASE] = 1
                    env_state[ENV_ID_ACTION] = (env_state[ENV_ROUND]) % 4

                    env_state[ENV_ROUND] += 1
                    new_sheriff = int((env_state[ENV_ROUND] - 1) % 4)
                    env_state[
                        ATTRIBUTE_PLAYER * new_sheriff + 2
                    ] = 1  # gán tư cách sheriff

                    if env_state[ENV_ID_ACTION] == (env_state[ENV_ROUND] - 1) % 4:
                        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4
                    env_state[ENV_NUMBER_CHECKED] = 0
                else:
                    env_state[ENV_PHASE] = 6
                    env_state[ENV_ID_ACTION] = (env_state[ENV_ROUND] - 1) % 4
        else:
            # không check
            env_state[ENV_LAST_CHECKED : ENV_LAST_CHECKED + NUMBER_PLAYER] = 0
            # B1: Cập nhật coin
            player_in4[P_COIN] += player_checked[P_COIN_BRIBE]
            player_checked[P_COIN_BRIBE] = 0
            # B3: cập nhật thẻ done sheriff and player_checked
            player_in4[-75:-60] = player_in4[-75:-60] + player_checked[-90:-75]
            player_checked[-90:-75] = np.zeros(NUMBER_TYPE_CARD)
            # cập nhật thẻ done từ hối lộ thẻ trong túi và cập nhật thẻ trong túi
            player_in4[-75:-60] = player_in4[-75:-60] + player_checked[-60:-45]
            player_checked[-60:-45] = np.zeros(15)
            # cập nhật thẻ done player_checked
            player_checked[-75:-60] = player_checked[-75:-60] + player_checked[-45:-30]
            player_checked[-45:-30] = np.zeros(NUMBER_TYPE_CARD)
            # cập nhật full thông tin
            player_checked[P_NUMBER_CARD_BRIBE_BAG] = 0
            player_checked[P_TYPE_IN_BAG:P_COIN_BRIBE] = 0
            player_checked[P_NUMBER_SMUGGLE] = np.sum(player_checked[-75:-60][4:])
            env_state[
                ATTRIBUTE_PLAYER * id_checked : ATTRIBUTE_PLAYER * (id_checked + 1)
            ] = player_checked
            env_state[
                ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
            ] = player_in4
            # cập nhật hệ thống lượt chơi
            if env_state[ENV_NUMBER_CHECKED] == 3:
                env_state[
                    ENV_TEMP_DROP : ENV_TEMP_DROP + NUMBER_TYPE_CARD * NUMBER_PLAYER
                ] = 0  # reset thẻ drop trong lượt
                env_state[ENV_PHASE] = 1

                old_sheriff = int((env_state[ENV_ROUND] - 1) % 4)
                env_state[ATTRIBUTE_PLAYER * old_sheriff + 2] = 0  # hủy tư cách sheriff

                env_state[ENV_ID_ACTION] = (env_state[ENV_ROUND]) % 4
                env_state[ENV_ROUND] += 1

                new_sheriff = int((env_state[ENV_ROUND] - 1) % 4)
                env_state[ATTRIBUTE_PLAYER * new_sheriff + 2] = 1  # gán tư cách sheriff

                if env_state[ENV_ID_ACTION] == (env_state[ENV_ROUND] - 1) % 4:
                    env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4
                env_state[ENV_NUMBER_CHECKED] = 0
            else:
                env_state[ENV_PHASE] = 6
                env_state[ENV_ID_ACTION] = (env_state[ENV_ROUND] - 1) % 4

    elif phase_env == 11:
        temp_card_drop = player_in4[-30:-15]
        card_up = env_state[
            ENV_LEFT_CARD
            + NUMBER_CARD_OPEN * (action - 80) : ENV_LEFT_CARD
            + NUMBER_CARD_OPEN * (action - 79)
        ]
        for i in range(len(temp_card_drop)):
            if temp_card_drop[i] > 0:
                card_up = np.append(np.array([i + 1] * int(temp_card_drop[i])), card_up)
        # if np.sum(temp_card_drop) > 0:
        card_up = card_up[: int(-np.sum(temp_card_drop))]
        temp_card_drop = np.zeros(NUMBER_TYPE_CARD)
        player_in4[-30:-15] = temp_card_drop
        env_state[
            ENV_LEFT_CARD
            + NUMBER_CARD_OPEN * (action - 80) : ENV_LEFT_CARD
            + NUMBER_CARD_OPEN * (action - 79)
        ] = card_up
        env_state[
            ATTRIBUTE_PLAYER * id_action : ATTRIBUTE_PLAYER * (id_action + 1)
        ] = player_in4
        env_state[ENV_LAST_CHECKED : ENV_LAST_CHECKED + NUMBER_PLAYER] = 0
        if env_state[ENV_NUMBER_CHECKED] == 3:
            env_state[ENV_PHASE] = 1
            # print('check phase 11')

            old_sheriff = int((env_state[ENV_ROUND] - 1) % 4)
            env_state[ATTRIBUTE_PLAYER * old_sheriff + 2] = 0  # hủy tư cách sheriff

            env_state[ENV_ID_ACTION] = (env_state[ENV_ROUND]) % 4
            env_state[ENV_ROUND] += 1

            new_sheriff = int((env_state[ENV_ROUND] - 1) % 4)
            env_state[ATTRIBUTE_PLAYER * new_sheriff + 2] = 1  # gán tư cách sheriff

            if env_state[ENV_ID_ACTION] == (env_state[ENV_ROUND] - 1) % 4:
                env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4
            env_state[ENV_NUMBER_CHECKED] = 0
        else:
            env_state[ENV_PHASE] = 6
            env_state[ENV_ID_ACTION] = (env_state[ENV_ROUND] - 1) % 4
    return env_state


@njit()
def check_winner(env_state):
    # mảng đếm số loại tài nguyên chính ngạch của 4 người chơi
    all_number_type_card = np.zeros(16)
    # mảng đếm hàng hóa buôn thành công của người chơi
    all_done_card = np.zeros(60)
    all_player_coin = np.array(
        [
            env_state[ATTRIBUTE_PLAYER * id + P_COIN]
            - env_state[ATTRIBUTE_PLAYER * id + P_DEBT]
            for id in range(4)
        ]
    )

    for id_player in range(4):
        player_i = env_state[
            ATTRIBUTE_PLAYER * id_player : ATTRIBUTE_PLAYER * (id_player + 1)
        ]
        player_i_done = player_i[-75:-60]  # thẻ hàng hóa buôn thành công của người chơi
        all_done_card[15 * id_player : 15 * (id_player + 1)] = player_i_done
        all_player_coin[id_player] = all_player_coin[id_player] + np.sum(
            np.multiply(player_i_done, ALL_REWARD)
        )
        all_done_card_i = player_i_done * ALL_NUMBER_COUNT
        # ghi nhận từng loại tài nguyên chính ngạch của người chơi
        all_number_type_card[id_player] = (
            all_done_card_i[0] + all_done_card_i[8] + all_done_card_i[12]
        )
        all_number_type_card[id_player + 4] = (
            all_done_card_i[1] + all_done_card_i[9] + all_done_card_i[13]
        )
        all_number_type_card[id_player + 8] = (
            all_done_card_i[2] + all_done_card_i[10] + all_done_card_i[14]
        )
        all_number_type_card[id_player + 12] = all_done_card_i[3] + all_done_card_i[11]
    reward_King = np.array([20, 15, 15, 10])
    reward_Queen = np.array([10, 10, 10, 5])
    for type_i in range(4):
        count_type = all_number_type_card[4 * type_i : 4 * (type_i + 1)]
        top1 = np.max(count_type)
        if top1 == 0:
            continue
        count_top1 = len(np.where(count_type == top1)[0])
        if count_top1 > 1:
            reward_i = (reward_King[type_i] + reward_Queen[type_i]) // count_top1
            all_player_coin = np.where(
                count_type == top1, all_player_coin + reward_i, all_player_coin
            )
        else:
            reward_i = reward_King[type_i]
            all_player_coin = np.where(
                count_type == top1, all_player_coin + reward_i, all_player_coin
            )
            count_type = np.where(count_type == top1, 0, count_type)
            top2 = np.max(count_type)
            if top2 == 0:
                continue
            else:
                count_top2 = len(np.where(count_type == top2)[0])
                if count_top2 > 1:
                    reward_i = reward_Queen[type_i] // count_top2
                    all_player_coin = np.where(
                        count_type == top2, all_player_coin + reward_i, all_player_coin
                    )
                else:
                    reward_i = reward_Queen[type_i]
                    all_player_coin = np.where(
                        count_type == top2, all_player_coin + reward_i, all_player_coin
                    )

    all_player_coin_int = np.array(
        [int(all_player_coin[i]) for i in range(len(all_player_coin))]
    )
    players_compare = np.zeros(NUMBER_PLAYER)
    temp_win = np.max(all_player_coin_int)
    number_win = np.where(all_player_coin_int == temp_win)[0]
    players_compare[number_win] = 1
    if len(number_win) == 1:
        return number_win[0]
    else:
        all_legal = np.zeros(4)
        for id in number_win:
            all_legal[id] = np.sum(all_done_card[15 * id : 15 * (id + 1)][:4])
        all_legal = all_legal * players_compare
        temp_win = np.max(all_legal)
        number_win = np.where(all_legal == temp_win)[0]
        if len(number_win) == 1:
            return number_win[0]
        else:
            players_compare = np.where(all_legal == temp_win, 1, 0)
            all_unlegal = np.zeros(4)
            for id in number_win:
                all_unlegal[id] = np.sum(all_done_card[15 * id : 15 * (id + 1)][4:])
            all_unlegal = players_compare * all_unlegal
            temp_win = np.max(all_unlegal)
            number_win = np.where(all_unlegal == temp_win)[0]
            if len(number_win) == 1:
                return number_win[0]
            else:
                all_player_coin = all_player_coin + np.array([0, 0.1, 0.2, 0.3])
                players_compare = np.where(all_unlegal == temp_win, 1, 0)
                number_win = np.where(players_compare == 1)[0]
                all_player_coin_last = np.zeros(4)
                all_player_coin_last[number_win] = all_player_coin[number_win]
                return np.argmax(all_player_coin_last)


@njit()
def getReward(player_state):
    check_end = player_state[P_CHECK_END]
    if check_end == 0:
        return -1
    elif check_end == 1 and player_state[P_ROUND] > 8:
        all_player_coin = np.array(
            [player_state[P_COIN] - player_state[P_DEBT]]
            + [
                (
                    player_state[
                        P_OTHER_PLAYER_IN4 + OTHER_ATTRIBUTE_PLAYER * id + P_COIN
                    ]
                    - player_state[
                        P_OTHER_PLAYER_IN4 + OTHER_ATTRIBUTE_PLAYER * id + P_DEBT
                    ]
                )
                for id in range(3)
            ]
        )
        all_player_order = player_state[P_ORDER : P_ORDER + NUMBER_PLAYER]
        all_player_coin = all_player_coin + all_player_order / 10
        all_number_type_card = np.zeros(16)
        all_done_card = np.zeros(60)
        for id_player in range(4):
            player_i_done = np.zeros(NUMBER_TYPE_CARD)
            if id_player == 0:
                player_i = player_state[:P_OTHER_PLAYER_IN4]
                player_i_done = player_i[
                    -75:-60
                ]  # thẻ hàng hóa buôn thành công của người chơi

            else:
                player_i_done = player_state[
                    P_OTHER_PLAYER_DONE_CARD
                    + NUMBER_TYPE_CARD * (id_player - 1) : P_OTHER_PLAYER_DONE_CARD
                    + NUMBER_TYPE_CARD * id_player
                ]
            all_done_card[15 * id_player : 15 * (id_player + 1)] = player_i_done
            all_player_coin[id_player] = all_player_coin[id_player] + np.sum(
                np.multiply(player_i_done, ALL_REWARD)
            )
            all_done_card_i = player_i_done * ALL_NUMBER_COUNT
            # ghi nhận từng loại tài nguyên chính ngạch của người chơi
            all_number_type_card[id_player] = (
                all_done_card_i[0] + all_done_card_i[8] + all_done_card_i[12]
            )
            all_number_type_card[id_player + 4] = (
                all_done_card_i[1] + all_done_card_i[9] + all_done_card_i[13]
            )
            all_number_type_card[id_player + 8] = (
                all_done_card_i[2] + all_done_card_i[10] + all_done_card_i[14]
            )
            all_number_type_card[id_player + 12] = (
                all_done_card_i[3] + all_done_card_i[11]
            )
        reward_King = np.array([20, 15, 15, 10])
        reward_Queen = np.array([10, 10, 10, 5])
        for type_i in range(4):
            count_type = all_number_type_card[4 * type_i : 4 * (type_i + 1)]
            top1 = np.max(count_type)
            if top1 == 0:
                continue
            count_top1 = len(np.where(count_type == top1)[0])
            if count_top1 > 1:
                reward_i = (reward_King[type_i] + reward_Queen[type_i]) // count_top1
                all_player_coin = np.where(
                    count_type == top1, all_player_coin + reward_i, all_player_coin
                )
            else:
                reward_i = reward_King[type_i]
                all_player_coin = np.where(
                    count_type == top1, all_player_coin + reward_i, all_player_coin
                )
                count_type = np.where(count_type == top1, 0, count_type)
                top2 = np.max(count_type)
                if top2 == 0:
                    continue
                else:
                    count_top2 = len(np.where(count_type == top2)[0])
                    if count_top2 > 1:
                        reward_i = reward_Queen[type_i] // count_top2
                        all_player_coin = np.where(
                            count_type == top2,
                            all_player_coin + reward_i,
                            all_player_coin,
                        )
                    else:
                        reward_i = reward_Queen[type_i]
                        all_player_coin = np.where(
                            count_type == top2,
                            all_player_coin + reward_i,
                            all_player_coin,
                        )

        all_player_coin_int = np.array(
            [int(all_player_coin[i]) for i in range(len(all_player_coin))]
        )
        players_compare = np.zeros(NUMBER_PLAYER)
        temp_win = np.max(all_player_coin_int)
        number_win = np.where(all_player_coin_int == temp_win)[0]
        players_compare[number_win] = 1
        if 0 not in number_win:
            return 0
        else:
            if len(number_win) == 1:
                return 1
            else:
                all_legal = np.zeros(4)
                for id in number_win:
                    all_legal[id] = np.sum(all_done_card[15 * id : 15 * (id + 1)][:4])
                all_legal = all_legal * players_compare
                temp_win = np.max(all_legal)
                number_win = np.where(all_legal == temp_win)[0]
                if 0 not in number_win:
                    return 0
                else:
                    if len(number_win) == 1:
                        return 1
                    else:
                        players_compare = np.where(all_legal == temp_win, 1, 0)
                        all_unlegal = np.zeros(4)
                        for id in number_win:
                            all_unlegal[id] = np.sum(
                                all_done_card[15 * id : 15 * (id + 1)][4:]
                            )
                        all_unlegal = players_compare * all_unlegal
                        temp_win = np.max(all_unlegal)
                        number_win = np.where(all_unlegal == temp_win)[0]
                        if 0 not in number_win:
                            return 0
                        else:
                            if len(number_win) == 1:
                                return 1
                            else:
                                players_compare = np.where(
                                    all_unlegal == temp_win, 1, 0
                                )
                                number_win = np.where(players_compare == 1)[0]
                                all_player_coin_last = np.zeros(4)
                                all_player_coin_last[number_win] = all_player_coin[
                                    number_win
                                ]
                                if np.argmax(all_player_coin_last) == 0:
                                    return 1
                                else:
                                    return 0
    else:
        return -1


@njit()
def bot_lv0(state, perData):
    validActions = getValidActions(state)
    arr_action = np.where(validActions == 1)[0]
    idx = np.random.randint(0, arr_action.shape[0])
    return arr_action[idx], perData


@njit()
def one_game_numba(p0, list_other, per_player, per1, per2, per3, p1, p2, p3):
    env_state = initEnv()
    count_turn = 0
    while system_check_end(env_state) and count_turn < 1000:
        # police_check = env_state[2:305:100]
        # print(police_check)
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
            if getValidActions(player_state)[action] != 1:
                raise Exception("bot dua ra action khong hop le")
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        else:
            raise Exception("Sai list_other.")
        env_state = stepEnv(env_state, action)
        count_turn += 1

    env_state[ENV_CHECK_END] = 1
    win = check_winner(env_state)
    for p_idx in range(4):
        env_state[ENV_PHASE] = 1
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)

        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        else:
            raise Exception("Sai list_other.")

        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4

    winner = False
    if np.where(list_other == -1)[0] == win:
        winner = True
    else:
        winner = False
    return winner, per_player


def one_game_normal(p0, list_other, per_player, per1, per2, per3, p1, p2, p3):
    env_state = initEnv()
    count_turn = 0
    while system_check_end(env_state) and count_turn < 1000:
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
            if getValidActions(player_state)[action] != 1:
                raise Exception("bot dua ra action khong hop le")
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        else:
            raise Exception("Sai list_other.")
        env_state = stepEnv(env_state, action)
        count_turn += 1

    env_state[ENV_CHECK_END] = 1

    for p_idx in range(4):
        env_state[ENV_PHASE] = 1
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        else:
            raise Exception("Sai list_other.")

        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % 4

    winner = False
    if np.where(list_other == -1)[0] == check_winner(env_state):
        winner = True
    else:
        winner = False
    return winner, per_player


@njit()
def n_games_numba(p0, num_game, per_player, list_other, per1, per2, per3, p1, p2, p3):
    win = 0
    for _ in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_numba(
            p0, list_other, per_player, per1, per2, per3, p1, p2, p3
        )
        win += winner

    return win, per_player


def n_games_normal(p0, num_game, per_player, list_other, per1, per2, per3, p1, p2, p3):
    win = 0
    for _ in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_normal(
            p0, list_other, per_player, per1, per2, per3, p1, p2, p3
        )
        win += winner

    return win, per_player


import importlib.util
import json, sys

try:
    from env import SHORT_PATH
except:
    pass


def load_module_player(player, game_name=None):
    if game_name == None:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/{player}/Agent_player.py"
        )
    else:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/ifelse/{game_name}/{player}.py"
        )
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


def load_agent(level, *args):
    num_bot = getAgentSize() - 1

    if "_level_" not in globals():
        global _level_
        _level_ = level
        init = True
    else:
        if _level_ != level:
            _level_ = level
            init = True
        else:
            init = False

    if init:
        global _list_per_level_
        global _list_bot_level_
        _list_per_level_ = []
        _list_bot_level_ = []

        if _level_ == 0:
            _list_per_level_ = [
                np.array([[0.0]], dtype=np.float64) for _ in range(num_bot)
            ]
            _list_bot_level_ = [bot_lv0 for _ in range(num_bot)]
        else:
            env_name = sys.argv[1]
            if len(args) > 0:
                dict_level = json.load(
                    open(f"{SHORT_PATH}src/Log/check_system_about_level.json")
                )
            else:
                dict_level = json.load(open(f"{SHORT_PATH}src/Log/level_game.json"))

            if str(_level_) not in dict_level[env_name]:
                raise Exception("Hiện tại không có level này")

            lst_agent_level = dict_level[env_name][str(level)][2]

            for i in range(num_bot):
                if level == -1:
                    module_agent = load_module_player(
                        lst_agent_level[i], game_name=env_name
                    )
                    _list_per_level_.append(module_agent.DataAgent())
                else:
                    data_agent_level = np.load(
                        f"{SHORT_PATH}src/Agent/{lst_agent_level[i]}/Data/{env_name}_{level}/Train.npy",
                        allow_pickle=True,
                    )
                    module_agent = load_module_player(lst_agent_level[i])
                    _list_per_level_.append(
                        module_agent.convert_to_test(data_agent_level)
                    )
                _list_bot_level_.append(module_agent.Test)

    return _list_bot_level_, _list_per_level_


@njit()
def check_run_under_njit(agent, perData):
    return True


def run(
    p0: any = bot_lv0,
    num_game: int = 100,
    per_player: np.ndarray = np.array([[0.0]]),
    level: int = 0,
    *args,
):
    num_bot = getAgentSize() - 1
    list_other = np.array([-1] + [i + 1 for i in range(num_bot)])
    try:
        check_njit = check_run_under_njit(p0, per_player)
    except:
        check_njit = False

    load_agent(level, *args)

    if check_njit:
        return n_games_numba(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
        )
    else:
        return n_games_normal(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
        )



import sys
import warnings

import numba
import numpy as np
from numba import jit, njit
from numba.core.errors import (
    NumbaDeprecationWarning,
    NumbaPendingDeprecationWarning,
    NumbaWarning,
)
from numba.typed import List

warnings.simplefilter("ignore", category=NumbaDeprecationWarning)
warnings.simplefilter("ignore", category=NumbaPendingDeprecationWarning)
warnings.simplefilter("ignore", category=NumbaWarning)


@njit()
def initEnv():
    env = np.zeros(81)
    card = np.arange(52)  # card
    np.random.shuffle(card)
    for i in range(4):
        env[card[i * 8 : (i + 1) * 8]] = i + 1
    env[52] = card[-1]  # trump suit card
    env[53] = 0  # mode: attack or defense
    env[54:57] = [1, 3, 4]
    env[57] = 1  # num of people choose attack this round
    env[58] = 2  # player_id defending
    env[59] = 0  # index player attack in env[54:57]
    env[60:80] = card[32:52]  # card on deck
    env[80] = 0
    return env


@njit()
def getStateSize():
    return 167


@njit()
def getAgentSize():
    return 4


@njit()
def getAgentState(env):
    state = np.zeros(getStateSize())
    if env[53] == 0:
        player_id = env[54:57][int(env[59])]  # attack player id
    elif env[53] == 1:
        player_id = env[58]  # defend player id
    state[0:52][np.where(env[0:52] == player_id)] = 1  #  card player hold
    state[52:104][np.where(env[0:52] == 6)] = 1  # all card defender defend this round
    state[104:156][np.where(env[0:52] == 5)] = 1  # card have to defend this round
    if env[53] == 1:
        state[156:158] = [0, 1]  #  attack, defend
    elif env[53] == 0:
        state[156:158] = [1, 0]
    state[158:162][int(env[52]) // 13] = 1  # trump suit
    state[162] = len(np.where(env[0:52] == 0)[0])  # num card on deck

    for i in range(3):
        state[163 + i] = np.where(env[0:52] == (player_id + i) % 4 + 1)[0].shape[0]

    state[166] = env[80]
    return state


@njit()
def getActionSize():
    return 53


@njit()
def getDefenseCard(state):
    card = np.zeros(52)
    idx = np.argmax(state[158:162])  # trump suit: 0:spade,1:club,2:diamond,3:heart
    card_def_id = np.argmax(state[104:156])
    if card_def_id // 13 != idx:  # card have to defend not a trump card
        card[13 * idx : 13 * (idx + 1)][
            np.where(state[13 * idx : 13 * (idx + 1)] == 1)
        ] = 1  #  trump card on hand
        card[card_def_id + 1 : 13 * (card_def_id // 13 + 1)][
            np.where(state[card_def_id + 1 : 13 * (card_def_id // 13 + 1)] == 1)
        ] = 1  # same type card, higher value on hand.
    else:  # card have to defend is a trump card
        card[card_def_id + 1 : 13 * (idx + 1)][
            np.where(state[card_def_id + 1 : 13 * (idx + 1)] == 1)
        ] = 1  # higher value trump card only.
    return card


@njit()
def getAttackCard(state):
    card = np.zeros(52)
    card_on_board = np.where(state[52:104] == 1)[0]
    card_value_on_board = card_on_board % 13  # value of that card (ex: 4 diamond is 4)
    card_on_hand = np.where(state[0:52] == 1)[0]  # card on player's hand
    for c in card_on_hand:
        if c % 13 in card_value_on_board:
            card[c] = 1
    return card


@njit()
def getValidActions(state):
    list_action = np.zeros(getActionSize())
    # attack
    if (
        state[156] == 1 and np.sum(state[52:104]) == 0
    ):  # main attacker, defender have nothing to defend yet.
        list_action[0:52] = state[0:52]
    elif (
        state[156] == 1 and np.sum(state[52:104]) != 0
    ):  # side attacker, attack only card with same value on the defend board( 4 heart on hand if have 4 spade on board)
        list_action[0:52] = getAttackCard(state)
        list_action[52] = 1
    # defense
    if state[157] == 1:  # defender
        list_action[0:52] = getDefenseCard(state)
        list_action[52] = 1
    if np.sum(list_action) == 0:
        list_action[52] = 1
    return list_action


@njit()
def drawCard(env):
    turn_draw_card = np.zeros(4)
    turn_draw_card[np.array([0, 2, 3])] = env[
        54:57
    ]  # attack player,main attack draw first.
    turn_draw_card[1] = env[58]  # defend player draw second.
    for p_id in turn_draw_card:  # draw card
        num_card_on_deck = len(np.where(env[0:52] == 0)[0])  # num cards left on deck
        if num_card_on_deck > 0:
            num_card_player = len(np.where(env[0:52] == p_id)[0])
            if num_card_player < 8:
                num_card_need = 8 - num_card_player
                if num_card_on_deck >= num_card_need:
                    env[
                        env[60:80].astype(np.int64)[
                            20
                            - num_card_on_deck : 20
                            - num_card_on_deck
                            + num_card_need
                        ]
                    ] = p_id
                else:
                    env[env[60:80].astype(np.int64)[20 - num_card_on_deck :]] = p_id
    return env


@njit()
def changeAttackPlayer(env):  # change the defender and attacker
    if env[58] == 1:
        env[54:57] = [4, 2, 3]
    elif env[58] == 2:
        env[54:57] = [1, 3, 4]
    elif env[58] == 3:
        env[54:57] = [2, 4, 1]
    elif env[58] == 4:
        env[54:57] = [3, 1, 2]


@njit()
def stepEnv(action, env):
    if action == 52:  # skip
        if env[53] == 1:  # defense
            env[0:52][np.where(env[0:52] == 5)] = env[58]  # Defender hold all card
            env[0:52][np.where(env[0:52] == 6)] = env[58]  # Defender hold all card
            env = drawCard(env)  # draw card
            env[58] = (
                (env[58] + 2) % 4 if env[58] > 2 else env[58] + 2
            )  # change defend player
            changeAttackPlayer(env)  # change attack player
            env[53] = 0  # reset mode: attack
            env[59] = 0
        elif env[53] == 0:  # attack
            env[57] += 1  # num attacker skip this round
            env[59] = (env[59] + 1) % 3
            if env[57] == 3:  # all attacker skip this round
                env[0:52][np.where(env[0:52] == 5)] = -1  # Thrown away card
                env[0:52][np.where(env[0:52] == 6)] = -1  # Thrown away card
                env = drawCard(env)  # draw card
                env[58] = 1 if env[58] == 4 else env[58] + 1  # change defend player
                changeAttackPlayer(env)  # change attack players
                env[57] = 0
                env[59] = 0

    else:  # attack or defend any card
        if env[53] == 1:  # defense
            env[0:52][np.where(env[0:52] == 5)] = 6  # defense this card successful
            env[action] = 6
            if len(np.where(env[0:52] == env[58])[0]) == 0:
                env[0:52][np.where(env[0:52] == 5)] = -1  # Thrown away card
                env[0:52][np.where(env[0:52] == 6)] = -1  # Thrown away card
                env = drawCard(env)  # draw card
                env[58] = 1 if env[58] == 4 else env[58] + 1  # change defend player
                changeAttackPlayer(env)  # change attack players
                env[57] = 0
                env[59] = 0
            env[53] = 0  # change mode: attack
        elif env[53] == 0:  # attack
            env[action] = 5  # this card have to defend
            env[53] = 1  # change mode: defend
            env[57] = 0  # change num player attack skip turn to 0
            env[59] = 0  # change attack player

    #  return env


@njit()
def checkEnded(env):
    if len(np.where(env[0:52] == 0)[0]) == 0:  # if no card left on deck
        for i in range(1, 5):
            if len(np.where(env[0:52] == i)[0]) == 0:
                return i - 1
    return -1


@njit()
def getReward(state):
    if state[162] == 0 and state[166] == 1:
        if np.sum(state[0:52]) == 0:
            return 1
        elif np.min(state[163:166]) == 0:
            return 0
        else:
            return -1
    else:
        return -1


@njit()
def one_game_numba(p0, pIdOrder, per_player, per1, per2, per3, p1, p2, p3):
    env = initEnv()
    winner = -1
    turn = 0
    while True:
        turn += 1
        if env[53] == 1:  # defend
            pIdx = int(env[58] - 1)
        else:  #  attack
            pIdx = int(env[54:57][int(env[59])] - 1)
        if pIdOrder[pIdx] == -1:
            action, per_player = p0(getAgentState(env), per_player)
        elif pIdOrder[pIdx] == 1:
            action, per1 = p1(getAgentState(env), per1)
        elif pIdOrder[pIdx] == 2:
            action, per2 = p2(getAgentState(env), per2)
        elif pIdOrder[pIdx] == 3:
            action, per3 = p3(getAgentState(env), per3)
        stepEnv(action, env)

        winner = checkEnded(env)
        if winner != -1:
            #  print(winner)
            break
    env[80] = 1
    #  env[0:52] = 6
    for pIdx in range(4):
        env[53] = 1
        env[58] = pIdx * 1.0 + 1.0
        if pIdOrder[pIdx] == -1:
            action, per_player = p0(getAgentState(env), per_player)
        elif pIdOrder[pIdx] == 1:
            action, per1 = p1(getAgentState(env), per1)
        elif pIdOrder[pIdx] == 2:
            action, per2 = p2(getAgentState(env), per2)
        elif pIdOrder[pIdx] == 3:
            action, per3 = p3(getAgentState(env), per3)

    win = False
    if np.where(pIdOrder == -1)[0][0] == checkEnded(env):
        win = True
    else:
        win = False

        #  #  print('ok')
    return win, per_player


@njit()
def n_game_numba(p0, num_game, per_player, list_other, per1, per2, per3, p1, p2, p3):
    win = 0
    for _n in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_numba(
            p0, list_other, per_player, per1, per2, per3, p1, p2, p3
        )
        #  print(winner)
        win += winner
    return win, per_player


import importlib.util
import json, sys

try:
    from env import SHORT_PATH
except:
    pass


def load_module_player(player, game_name=None):
    if game_name == None:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/{player}/Agent_player.py"
        )
    else:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/ifelse/{game_name}/{player}.py"
        )
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


def one_game_normal(p0, pIdOrder, per_player, per1, per2, per3, p1, p2, p3):
    env = initEnv()
    winner = -1
    turn = 0
    while True:
        turn += 1
        if env[53] == 1:
            pIdx = int(env[58] - 1)
        else:
            pIdx = int(env[54:57][int(env[59])] - 1)
        if pIdOrder[pIdx] == -1:
            action, per_player = p0(getAgentState(env), per_player)
        elif pIdOrder[pIdx] == 1:
            action, per1 = p1(getAgentState(env), per1)
        elif pIdOrder[pIdx] == 2:
            action, per2 = p2(getAgentState(env), per2)
        elif pIdOrder[pIdx] == 3:
            action, per3 = p3(getAgentState(env), per3)
        stepEnv(action, env)

        winner = checkEnded(env)
        if winner != -1:
            break
    env[80] = 1
    #  env[0:52] = 6
    for pIdx in range(4):
        env[53] = 1
        env[58] = pIdx * 1.0 + 1.0
        if pIdOrder[pIdx] == -1:
            action, per_player = p0(getAgentState(env), per_player)
        elif pIdOrder[pIdx] == 1:
            action, per1 = p1(getAgentState(env), per1)
        elif pIdOrder[pIdx] == 2:
            action, per2 = p2(getAgentState(env), per2)
        elif pIdOrder[pIdx] == 3:
            action, per3 = p3(getAgentState(env), per3)

    win = False
    if np.where(pIdOrder == -1)[0][0] == checkEnded(env):
        win = True
    else:
        win = False

        #  #  print('ok')
    return win, per_player


def n_game_normal(p0, num_game, per_player, list_other, per1, per2, per3, p1, p2, p3):
    win = 0
    for _n in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_normal(
            p0, list_other, per_player, per1, per2, per3, p1, p2, p3
        )
        #  print(winner)
        win += winner
    return win, per_player


@njit()
def bot_lv0(state, perData):
    validActions = getValidActions(state)
    arr_action = np.where(validActions == 1)[0]
    idx = np.random.randint(0, arr_action.shape[0])
    return arr_action[idx], perData


def load_agent(level, *args):
    num_bot = getAgentSize() - 1

    if "_level_" not in globals():
        global _level_
        _level_ = level
        init = True
    else:
        if _level_ != level:
            _level_ = level
            init = True
        else:
            init = False

    if init:
        global _list_per_level_
        global _list_bot_level_
        _list_per_level_ = []
        _list_bot_level_ = []

        if _level_ == 0:
            _list_per_level_ = [
                np.array([[0.0]], dtype=np.float64) for _ in range(num_bot)
            ]
            _list_bot_level_ = [bot_lv0 for _ in range(num_bot)]
        else:
            env_name = sys.argv[1]
            if len(args) > 0:
                dict_level = json.load(
                    open(f"{SHORT_PATH}src/Log/check_system_about_level.json")
                )
            else:
                dict_level = json.load(open(f"{SHORT_PATH}src/Log/level_game.json"))

            if str(_level_) not in dict_level[env_name]:
                raise Exception("Hiện tại không có level này")

            lst_agent_level = dict_level[env_name][str(level)][2]

            for i in range(num_bot):
                if level == -1:
                    module_agent = load_module_player(
                        lst_agent_level[i], game_name=env_name
                    )
                    _list_per_level_.append(module_agent.DataAgent())
                else:
                    data_agent_level = np.load(
                        f"{SHORT_PATH}src/Agent/{lst_agent_level[i]}/Data/{env_name}_{level}/Train.npy",
                        allow_pickle=True,
                    )
                    module_agent = load_module_player(lst_agent_level[i])
                    _list_per_level_.append(
                        module_agent.convert_to_test(data_agent_level)
                    )
                _list_bot_level_.append(module_agent.Test)

    return _list_bot_level_, _list_per_level_


@njit()
def check_run_under_njit(agent, perData):
    return True


def run(
    p0: any = bot_lv0,
    num_game: int = 100,
    per_player: np.ndarray = np.array([[0.0]]),
    level: int = 0,
    *args,
):
    num_bot = getAgentSize() - 1
    list_other = np.array([-1] + [i + 1 for i in range(num_bot)])
    try:
        check_njit = check_run_under_njit(p0, per_player)
    except:
        check_njit = False

    load_agent(level, *args)

    if check_njit:
        return n_game_numba(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
        )
    else:
        return n_game_normal(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
        )



import warnings
import env
from src.Utils import load_module_player

warnings.filterwarnings("ignore")
from numba.core.errors import (
    NumbaDeprecationWarning,
    NumbaExperimentalFeatureWarning,
    NumbaPendingDeprecationWarning,
    NumbaWarning,
)

warnings.simplefilter("ignore", category=NumbaDeprecationWarning)
warnings.simplefilter("ignore", category=NumbaPendingDeprecationWarning)
warnings.simplefilter("ignore", category=NumbaExperimentalFeatureWarning)
warnings.simplefilter("ignore", category=NumbaWarning)

COUNT_TEST = 1000


# check hết hệ thống
def CheckAllFunc(Agent_name, BOOL_CHECK_ENV, msg):
    env.make("SushiGo")
    Agent = load_module_player(Agent_name)
    for func in ["DataAgent", "Train", "Test", "convert_to_save", "convert_to_test"]:
        try:
            getattr(Agent, func)
        except:
            msg.append(f"Không có hàm: {func}")
            BOOL_CHECK_ENV = False
    return BOOL_CHECK_ENV, msg


def CheckRunGame(Agent_name, BOOL_CHECK_ENV, msg):
    for game_name in [
        "Splendor_v3",
        "MachiKoro",
        "SushiGo",
    ]:
        env.make(game_name)
        Agent = load_module_player(Agent_name)
        try:
            per = Agent.DataAgent()
            win, per = env.run(Agent.Train, COUNT_TEST, per, 0)
        except:
            msg.append(f"Train đang bị lỗi {game_name}")
            BOOL_CHECK_ENV = False
            break

        try:
            per = Agent.convert_to_test(Agent.convert_to_save(Agent.DataAgent()))
            win, per = env.run(Agent.Test, COUNT_TEST, per, 0)
        except:
            msg.append(f"Test đang bị lỗi {game_name}")
            BOOL_CHECK_ENV = False
            break

    return BOOL_CHECK_ENV, msg


def check_agent(Agent_name):
    BOOL_CHECK_ENV = True
    msg = []
    print(
        Agent_name,
        "| Function checking ...",
    )
    BOOL_CHECK_ENV, msg = CheckAllFunc(Agent_name, BOOL_CHECK_ENV, msg)
    print(
        Agent_name,
        "| Run game checking ...",
    )
    BOOL_CHECK_ENV, msg = CheckRunGame(Agent_name, BOOL_CHECK_ENV, msg)
    return BOOL_CHECK_ENV, msg



from setuptools import setup, find_packages
from os import path
# twine upload dist/*
# python setup.py bdist_wheel
classifiers = [
  'Development Status :: 5 - Production/Stable',
  'Intended Audience :: Education',
  'Operating System :: Microsoft :: Windows :: Windows 10',
  'License :: OSI Approved :: MIT License',
  'Programming Language :: Python :: 3.6',
  'Programming Language :: Python :: 3.7',
  'Programming Language :: Python :: 3.8',
  'Programming Language :: Python :: 3.9',
]
 
setup(
  name='vis-lab',
  version='1.0.2',
  description='a system reinforcement learning game splendor for vis intern',
  # long_description=open(path.join(path.abspath(path.dirname(__file__)), 'README.md'), encoding='utf-8').read(),
  url='',  
  author='Ngo Xuan Phong',
  author_email='phong@cht.edu.vn',
  license='MIT', 
  classifiers=classifiers,
  keywords='python',
  packages=find_packages(),
  install_requires=[
    'numpy == 1.21.6',
    'numba == 0.56.3',
    'llvmlite == 0.39.1',
    'setuptools == 57.4.0',
    'importlib-metadata == 4.13.0',
    'typing-extensions == 4.1.1',
    'zipp == 3.10.0'
    ],
  python_requires='>=3.6',
  include_package_data=True
)


import warnings


warnings.filterwarnings("ignore")
from numba.core.errors import (
    NumbaDeprecationWarning,
    NumbaExperimentalFeatureWarning,
    NumbaPendingDeprecationWarning,
    NumbaWarning,
)

warnings.simplefilter("ignore", category=NumbaDeprecationWarning)
warnings.simplefilter("ignore", category=NumbaPendingDeprecationWarning)
warnings.simplefilter("ignore", category=NumbaExperimentalFeatureWarning)
warnings.simplefilter("ignore", category=NumbaWarning)

import os


def check_graphics(game_name):
    BOOL_CHECK_GRAPHICS = True
    msg = []
    if os.path.exists(f"src/Base/{game_name}/_render_func.py") == False:
        BOOL_CHECK_GRAPHICS = False
        msg.append("_render_func.py not found")
    elif os.path.exists(f"src/Base/{game_name}/env.py") == False:
        BOOL_CHECK_GRAPHICS = False
        msg.append("env.py not found")
    return BOOL_CHECK_GRAPHICS, msg



#  state based
#  small NN deep
import sys
import numpy as np
from numba import njit
import env

game_name = sys.argv[1]
env.make(game_name)


from numba.typed import List


def DataAgent():
    per = List(
        [
            np.zeros(
                (100, env.getStateSize(), env.getActionSize())
            ),  # [0][value][idS] khi mở đầu game
            np.zeros(
                (100, env.getStateSize(), env.getActionSize())
            ),  # [1][value][ids] lưu lại cuối cùng
            np.zeros((1, 1, 1)),  # [2][0][0][0] vừa thắng hay thua
        ]
    )
    return per


@njit()
def Train(state, per):
    actions = env.getValidActions(state)
    weights = np.zeros(env.getActionSize())
    if per[2][0][0][0] == 0:
        temp = np.arange(env.getActionSize(), dtype=np.float64)
        np.random.shuffle(temp)
        weights += temp
        for ids in range(env.getStateSize()):
            if state[ids] < 100:
                per[0][int(state[ids])][ids] += temp / np.max(temp)
    else:
        for ids in range(env.getStateSize()):
            if state[ids] < 100:
                weights += (
                    np.argsort(np.argsort(per[0][int(state[ids])][ids]))
                    / env.getActionSize()
                )
    output = weights * actions + actions
    action = np.argmax(output)
    win = env.getReward(state)
    if win != -1:
        if win == 1:
            #  print("đây")
            per[1] += per[0]
            per[2][0][0][0] = 1
        else:
            per[0][:, :, :] = 0.0
            per[2][0][0][0] = 0
    return action, per


@njit
def Test(state, per):
    state_int = state.astype(np.int64)
    stateSize = env.getStateSize()
    actionSize = env.getActionSize()
    where_ = np.where((state_int < per[stateSize][0]) & (state_int >= 0))[0]
    weight = np.zeros(actionSize)
    for i in where_:
        weight += per[i][state_int[i]]

    actions = env.getValidActions(state)
    output = (weight + 1) * actions
    #  action = np.argmax(output)
    list_action = np.where(actions == 1)[0]
    action = list_action[np.argmax(output[list_action])]
    return action, per


def convert_to_save(perData):
    if len(perData) == env.getStateSize() + 1:
        raise Exception("Data này đã được convert rồi.")
    data = List()
    arr = np.zeros(env.getStateSize(), int)
    for i in range(env.getStateSize()):
        for j in range(100):
            if (perData[1][j, i] == 0).all():
                check = True
                for k in range(j, 100):
                    if (perData[1][k, i] != 0).any():
                        check = False
                        break
                if check:
                    arr[i] = j
                    break
        else:
            arr[i] = 100

    for i in range(env.getStateSize()):
        data.append(perData[1][: arr[i], i])
        for j in range(data[i].shape[0]):
            data[i][j] = np.argsort(np.argsort(data[i][j])) / float(env.getActionSize())

    data.append(np.array([arr], float))
    return data


def convert_to_test(perData):
    return List(perData)



import numpy as np
from PIL import Image, ImageDraw, ImageEnhance, ImageFont

from src.Base.GoFish import env as _env
from env import SHORT_PATH

IMG_PATH = SHORT_PATH + "src/Base/GoFish/images/"
BG_SIZE = (1680, 720)
CARD_SIZE = (100, 140)

CARD_SIZE1 = (60, 84)

CARD_SIZE2 = (70, 98)

CARD_SIZE3 = (50, 70)
card_values = ["A", "2", "3", "4", "5", "6", "7", "8", "9", "10", "J", "Q", "K"]


class Sprites:
    def __init__(self) -> None:
        self.background = (
            Image.open(IMG_PATH + "backGroundBlack.jpg").resize(BG_SIZE).convert("RGBA")
        )
        self.cards = []
        for value in card_values:
            self.cards.append(
                Image.open(IMG_PATH + f"{value}-Heart.png").convert("RGBA")
            )

        self.cards_ = []
        for i in self.cards:
            br_ = ImageEnhance.Brightness(i)
            self.cards_.append(br_.enhance(0.5))

        self.card_back = Image.open(IMG_PATH + "Card_back.png").convert("RGBA")
        br = ImageEnhance.Brightness(self.card_back)
        self.card_back_ = br.enhance(0.5)


sprites = Sprites()


def get_state_image(
    state=None,
):  ###-------------------------------------------------------
    state = state.astype(int)
    fontMin = ImageFont.FreeTypeFont("src/ImageFonts/arial.ttf", 20)
    font = ImageFont.FreeTypeFont("src/ImageFonts/arial.ttf", 35)
    font50 = ImageFont.FreeTypeFont("src/ImageFonts/arial.ttf", 45)
    bg = sprites.background.copy()
    drawler = ImageDraw.ImageDraw(bg)
    if state is None:
        return bg

    pl = 0
    if np.sum(state[64:67]) == 1:
        pl = np.where(state[64:67])[0][0] + 1

    # Phần bài của tôi
    # thor = Image.open(IMG_PATH+"thor.png").resize((75, 64))
    # bg.paste(thor, (500, 600))

    myCards = state[:13]
    my_cards = np.where((myCards < 4) & (myCards > 0))[0]
    point = state[14]

    if point:
        _ = 840
        for i in np.where(myCards == 4)[0]:
            card = sprites.cards[i].resize(CARD_SIZE1)
            _ += 15
            bg.paste(card, (_, 620))

    drawler.text((600, 640), "Score: " + str(point), (255, 0, 0), font50)

    n = my_cards.size
    for i in range(n):
        id = my_cards[i]
        laBai = sprites.cards[id].resize(CARD_SIZE)
        x = 820 - (50 + (n - 1) * 25) + i * 50
        bg.paste(laBai, (x, 470))
        num = state[id]
        drawler.text((x + 10, 560), str(num), (0, 0, 255), font)

    # Lá còn lại để bốc
    card_back = sprites.card_back.resize((80, 112))
    bg.paste(card_back, (1180, 40))
    card_back = sprites.card_back.resize((80, 112))
    bg.paste(card_back, (1175, 45))
    drawler.text((1270, 60), str(state[60]), (255, 255, 255), font50)
    if np.where(state[61:64])[0].size:
        phase = np.where(state[61:64])[0][0]
        drawler.text((100, 60), "Phase: " + str(phase), (255, 255, 255), font50)

    # Phần bài người chơi khác
    ###player 2
    if pl != 0 and pl != 2:
        card_back = sprites.card_back_.resize(CARD_SIZE)
    else:
        card_back = sprites.card_back.resize(CARD_SIZE)
    bg.paste(card_back, (710, 40))
    # drawler.text((450, 100), "Player 2:", (255, 255, 255), font50)
    drawler.text((725, 125), str(state[43]), (255, 255, 31), font50)
    drawler.text((500, 110), "Score: " + str(state[44]), (255, 255, 255), font50)
    point2 = state[44]
    if point2:
        cards = state[30:43]
        _ = 810
        for i in np.where(cards)[0]:
            if pl != 0 and pl != 2:
                card = sprites.cards_[i].resize(CARD_SIZE3)
            else:
                card = sprites.cards[i].resize(CARD_SIZE3)
            _ += 10
            bg.paste(card, (_, 117))

    cardYc_ = state[80:93]
    cardYc = np.where(cardYc_)[0]
    __ = 800
    for i in range(cardYc.size):
        id = cardYc[i]
        if pl != 0 and pl != 2:
            laBai = sprites.cards_[id].resize(CARD_SIZE2)
        else:
            laBai = sprites.cards[id].resize(CARD_SIZE2)
        __ += 20
        bg.paste(laBai, (__, 10))

    ###player 1---------------
    if pl != 0 and pl != 1:
        card_back = sprites.card_back_.resize(CARD_SIZE)
    else:
        card_back = sprites.card_back.resize(CARD_SIZE)
    bg.paste(card_back, (220, 250))
    # drawler.text((450, 100), "Player 2:", (255, 255, 255), font50)
    drawler.text((235, 335), str(state[28]), (255, 255, 31), font50)
    drawler.text((15, 320), "Score: " + str(state[29]), (255, 255, 255), font50)
    point2 = state[29]
    if point2:
        cards = state[15:28]
        _ = 320
        for i in np.where(cards)[0]:
            if pl != 0 and pl != 2:
                card = sprites.cards_[i].resize(CARD_SIZE3)
            else:
                card = sprites.cards[i].resize(CARD_SIZE3)
            _ += 10
            bg.paste(card, (_, 327))

    cardYc_ = state[67:80]
    cardYc = np.where(cardYc_)[0]
    __ = 310
    for i in range(cardYc.size):
        id = cardYc[i]
        if pl != 0 and pl != 1:
            laBai = sprites.cards_[id].resize(CARD_SIZE2)
        else:
            laBai = sprites.cards[id].resize(CARD_SIZE2)
        __ += 20
        bg.paste(laBai, (__, 220))

    ###player 3---------------
    if pl != 0 and pl != 3:
        card_back = sprites.card_back_.resize(CARD_SIZE)
    else:
        card_back = sprites.card_back.resize(CARD_SIZE)
    bg.paste(card_back, (1680 - 330, 250))
    # drawler.text((450, 100), "Player 2:", (255, 255, 255), font50)
    drawler.text((1680 - 313, 335), str(state[58]), (255, 255, 31), font50)
    drawler.text((1680 - 215, 320), "Score: " + str(state[59]), (255, 255, 255), font50)
    point2 = state[59]
    if point2:
        cards = state[45:58]
        _ = 1680 - 380
        for i in np.where(cards)[0]:
            if pl != 0 and pl != 2:
                card = sprites.cards_[i].resize(CARD_SIZE3)
            else:
                card = sprites.cards[i].resize(CARD_SIZE3)
            _ -= 10
            bg.paste(card, (_, 327))

    cardYc_ = state[93:106]
    cardYc = np.where(cardYc_)[0]
    __ = 1680 - 390
    for i in range(cardYc.size):
        id = cardYc[i]
        if pl != 0 and pl != 3:
            laBai = sprites.cards_[id].resize(CARD_SIZE2)
        else:
            laBai = sprites.cards[id].resize(CARD_SIZE2)
        __ -= 20
        bg.paste(laBai, (__, 220))

    return bg


def get_description(action):  # -----------------------------------------
    if action < 0 or action >= _env.getActionSize():
        return ""

    if action == 0:
        return "Bốc"
    elif action < 4:
        return "Yêu cầu player: " + str(action)
    else:
        return "Yêu cầu lá bài: " + str(card_values[action - 4])


class Env_components:
    def __init__(self, env, winner, list_other) -> None:
        self.env = env
        self.winner = winner
        self.list_other = list_other


def get_env_components():  # -------------------------------------------------
    env = _env.initEnv()
    winner = _env.checkEnded(env)
    list_other = np.array([-1, 1, 2, 3])
    np.random.shuffle(list_other)

    env_components = Env_components(env, winner, list_other)
    return env_components


def get_main_player_state(
    env_components: Env_components, list_agent, list_data, action=None
):  # ---------------------------------
    if not action is None:
        _env.stepEnv(action, env_components.env)

    env_components.winner = _env.checkEnded(env_components.env)
    if env_components.winner == -1:
        env = env_components.env.copy()
        while env[113] < 400:
            p_idx = int(env_components.env[113]) % 4
            if env_components.list_other[p_idx] == -1:
                break

            state = _env.getAgentState(env_components.env)
            agent = list_agent[env_components.list_other[p_idx] - 1]
            data = list_data[env_components.list_other[p_idx] - 1]
            action, data = agent(state, data)
            _env.stepEnv(action, env_components.env)

            env_components.winner = _env.checkEnded(env_components.env)
            if env_components.winner != -1:
                break

    if env_components.winner == -1:
        state = _env.getAgentState(env_components.env)
        win = -1
    else:
        my_idx = np.where(env_components.list_other == -1)[0][0]
        env = env_components.env.copy()
        env[113] = my_idx
        state = _env.getAgentState(env)
        if my_idx == env_components.winner:
            win = 1
        else:
            win = 0

        # Chạy turn cuối cho 3 bot hệ thống
        for p_idx in range(4):
            if p_idx != my_idx:
                env[113] = p_idx
                _state = _env.getAgentState(env)
                agent = list_agent[env_components.list_other[p_idx] - 1]
                data = list_data[env_components.list_other[p_idx] - 1]
                action, data = agent(_state, data)

    return win, state, env_components
